/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */


package com.funambol.framework.core;

import java.util.*;


/**
 * Thic class corresponds to the &lt;Sync&gt; command in the SyncML represent DTD
 *
 *
 *
 * @version $Id: Sync.java,v 1.2 2006/08/07 21:05:28 nichele Exp $
 */
public final class Sync
extends AbstractCommand
implements java.io.Serializable {

    // --------------------------------------------------------------- Constants
    public static String COMMAND_NAME = "Sync";

    // ------------------------------------------------------------ Private data
    private Target target;
    private Source source;
    private Meta meta;
    private ArrayList commands = new ArrayList();
    private int numberOfChanges;

    // ------------------------------------------------------------ Constructors

    /** For serialization purposes */
    protected Sync(){}

    /**
     * Creates a new Sync object
     *
     * @param cmdID the command identifier - NOT NULL
     * @param noResp is <b>true</b> if no response is required
     * @param cred the authentication credential
     * @param target the target object
     * @param source the source object
     * @param meta the meta object
     * @param numberOfChanges the number of changes
     * @param commands an array of elements that must be of one of the
     *                 following types: {@link Add}, {@link Atomic},
     *                 {@link Copy}, {@link Delete}, {@link Replace},
     *                 {@link Sequence}
     *
     * @throws java.lang.IllegalArgumentException
     *
     */
    public Sync(final CmdID cmdID,
                final boolean noResp,
                final Cred cred,
                final Target target,
                final Source source,
                final Meta meta,
                final int numberOfChanges,
                final AbstractCommand[] commands) {
        super(cmdID);

        setCommands(commands);
        setCred(cred);
        setMeta(meta);

        this.noResp  = (noResp) ? new Boolean(noResp) : null;
        this.target = target;
        this.source = source;
        this.numberOfChanges  = numberOfChanges;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Gets the Target object property
     *
     * @return target the Target object property
     */
    public Target getTarget() {
        return target;
    }

    /**
     * Sets the Target object property
     *
     * @param target the Target object property
     *
     */
    public void setTarget(Target target) {
        this.target = target;
    }

    /**
     * Gets the Source object property
     *
     * @return source the Source object property
     */
    public Source getSource() {
        return source;
    }

    /**
     * Gets the Source object property
     *
     * @param source the Source object property
     */
    public void setSource(Source source) {
        this.source = source;
    }

    /**
     *
     * @return The return value is guaranteed to be non-null.
     *          The array elements are guaranteed to be non-null.
     *
     */
    public ArrayList getCommands() {
        return this.commands;
    }

    /**
     * Sets the sequenced commands. The given commands must be of the allowed
     * types.
     *
     * @param commands the commands - NOT NULL and o the allawed types
     *
     * @throws IllegalArgumentException if the constraints are not met
     */
    public void setCommands(AbstractCommand[] commands) {
        if (commands == null) {
            throw new IllegalArgumentException("commands cannot be null");
        }

        for (int i = 0; i < commands.length; i++) {
            if (commands[i] == null) {
                throw new IllegalArgumentException("commands[" + i +"] cannot be null");
            } else if ((!(commands[i] instanceof Add))
                       && (!(commands[i] instanceof Replace))
                       && (!(commands[i] instanceof Delete))
                       && (!(commands[i] instanceof Copy))
                       && (!(commands[i] instanceof Atomic))
                       && (!(commands[i] instanceof Map))
                       && (!(commands[i] instanceof Sync))) {
                throw new IllegalArgumentException(
                    "commands[" + i + "] cannot be a " + commands[i].getName()
                );
            }
        }
        this.commands.clear();
        this.commands.addAll(Arrays.asList(commands));
    }

    /**
     * Gets the total number of changes
     *
     * @return the total number of changes
     */
    public int getNumberOfChanges() {
        return numberOfChanges;
    }

    /**
     * Sets the numberOfChanges property
     *
     * @param numberOfChanges the total number of changes
     */
    public void setNumberOfChanges(int numberOfChanges) {
        this.numberOfChanges = numberOfChanges;
    }

    public String getName() {
        return Sync.COMMAND_NAME;
    }
}
