/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.security.jaas;

import java.io.*;
import java.util.logging.Logger;
import java.util.logging.Level;

import javax.security.auth.callback.*;

import com.funambol.framework.core.Cred;
import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.tools.Base64;
import org.apache.commons.lang.StringUtils;


/**
 * This handler implements the JAAS <i>CallbackHandler</i> interface. It stores
 * a <i>Credential</i> object for later use as principal and credentials provider.
 * This simple implementation supports basic authentication stored bas64 encoded
 * in the form login:password.
 *
 * TO DO: supports MD5 authentication
 *
 *.com
 */
public class CredentialHandler implements CallbackHandler {

    // --------------------------------------------------------------- Constants

    public static final String TYPE_BASIC = "syncml:auth-basic";
    public static final String TYPE_CLEAR = "syncml:auth-clear";
    public static final String SUPPORTED_TYPES = TYPE_BASIC + ',' + TYPE_CLEAR;

    // ------------------------------------------------------------ Private data

    private String     login      = null;
    private char[]     password   = null;

    // ---------------------------------------------------------- Protected data

    protected final Logger log = FunambolLogger.getLogger();

    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new instance of CredentialHandler
     *
     * @param credential
     *
     * @throws IllegalArgumentException
     */
    public CredentialHandler(Cred credential) throws IllegalArgumentException {

        String type = credential.getType();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("credential: " + credential);
        }

        if (SUPPORTED_TYPES.indexOf(type) < 0) {
            throw new IllegalArgumentException( "Authorization type '"
                                              + type
                                              + "' not supported"
                                              );
        }

        if (TYPE_BASIC.equals(type)) {
            String s = new String(Base64.decode(credential.getData()));

            int p = s.indexOf(':');

            if (p == -1) {
                login = s;
                password = null;
            } else {
                login = (p>0) ? s.substring(0, p) : "";
                password = toChars((p == (s.length()-1)) ? "" : s.substring(p+1));
            }
        } else if (TYPE_CLEAR.equals(type)) {
            String s = credential.getData();

            int p = s.indexOf(':');

            if (p == -1) {
                login = s;
                password = null;
            } else {
                login = (p>0) ? s.substring(0, p) : "";
                password = toChars((p == (s.length()-1)) ? "" : s.substring(p+1));
            }
        }
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Returns the login of this credential.
     *
     * @return the login value
     */
    public String getLogin() {
        return login;
    }

    // ------------------------------------------------------------------ handle

    public void handle(Callback[] callbacks)
    throws IOException, UnsupportedCallbackException {
        for (int i = 0; i < callbacks.length; i++) {
            log.finest("Handling " + callbacks[i]);

            if (callbacks[i] instanceof TextOutputCallback) {

                // display the message according to the specified type
                TextOutputCallback toc = (TextOutputCallback)callbacks[i];
                switch (toc.getMessageType()) {
                   case TextOutputCallback.INFORMATION:
                      log.info(toc.getMessage());
                      break;
                   case TextOutputCallback.ERROR:
                      log.severe(toc.getMessage());
                      break;
                   case TextOutputCallback.WARNING:
                      log.warning(toc.getMessage());
                     break;
                   default:
                      throw new IOException( "Unsupported message type: "
                                           + toc.getMessageType()
                                           );
                }

            } else if (callbacks[i] instanceof NameCallback) {
                NameCallback nc = (NameCallback)callbacks[i];
                nc.setName(login);
            } else if (callbacks[i] instanceof PasswordCallback) {
                PasswordCallback nc = (PasswordCallback)callbacks[i];
                nc.setPassword(password);
            } else {
                throw new UnsupportedCallbackException
                 (callbacks[i], "Unrecognized Callback");
            }
          }
    }

    // --------------------------------------------------------- Private methods

    private char[] toChars(String str) {
        if (StringUtils.isEmpty(str)) {
            return new char[0];
        }

        int l = str.length();
        char[] ret = new char[l];

        str.getChars(0, l, ret, 0);

        return ret;
    }

}
