/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.transport.http;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class implements a <i>SyncHolder</i> cache. <br>
 * It holds an HashMap in its instance parameter <i>cache</i>.
 * <p>
 * An optimization performed in this object is about avoiding memory eating
 * due to missing SyncHolders removing. This problem is solved with the
 * following actions:
 * <ul>
 *   <li> each SyncHolder stores a creation timestamp that is set in the
 *        constructor to the current system time millis;
 *   <li> HolderCache stores in the lastCleaningTimestamp static field the last
 *        time the cache was cleared;
 *   <li> If the the difference between the current time millis and the
 *        lastCleaningTimestamp is greater then the configured sessionTimeToLive,
 *        the cleaner process is executed;
 *   <li> the cleaner process iterates the sync holders in the cache and closes and removes the expired ones;
 * </ul>
 * <p>
 * NOTES:
 * <p>
 * 1. The cleaning process could be moved to a separate thread and executed
 *    asynchronously respect to the execution of servlet execution.
 *
 *
 * @version $Id: SyncHolderCache.java,v 1.3 2006/08/07 21:09:23 nichele Exp $
 */
public class SyncHolderCache {

    // --------------------------------------------------------------- Constants

    public static final long   CLEANING_PERIOD = 300000; // 5 mins
    public static final long   DEFAULT_TTL     = 300000; // 5 mins
    public static final String LOG_NAME        = "funambol.transport.http";

    private static final Logger log = Logger.getLogger(LOG_NAME);

    // ------------------------------------------------------------ Private data

    private long holderTimeToLive = DEFAULT_TTL;

    private long lastCleaningTimestamp = 0;

    private Map cache;

    // ------------------------------------------------------------ Constructors


    public SyncHolderCache(long timeToLive) {
        cache = new HashMap(97);

        lastCleaningTimestamp = System.currentTimeMillis();

        holderTimeToLive = timeToLive;
    }

    // ------------------------------------------- Implementation of HolderCache

    public void put(SyncHolder holder) {
        clean();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Caching " + holder.getSessionId() + '(' + holder + ')');
        }

        synchronized (cache) {
            cache.put(holder.getSessionId(), holder);
        }
    }

    public SyncHolder get(String sessionId) {
        return (SyncHolder)cache.get(sessionId);
    }

    public void remove(String sessionId) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Removing holder for " + sessionId);
        }
        synchronized (cache) {
            cache.remove(sessionId);
        }
    }

    // ---------------------------------------------------- Other public methods

    public String toString() {
        String s = null;
        synchronized (cache) {
            s = String.valueOf(cache);
        }
        return s;
    }

    // --------------------------------------------------------- Private methods

    /**
     * If System.currentTimeMillis() - lastCleaningTimestamp is greater than
     * holderTimeToLive, purge old holders.
     *
     * @returns true if cleaning was performed, false otherwise
     */
    private boolean clean() {
        long now = System.currentTimeMillis();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Cleaning procedure..."                          );
            log.finest("now: "                   + now                  );
            log.finest("CLEANING_PERIOD: "       + CLEANING_PERIOD      );
            log.finest("lastCleaningTimestamp: " + lastCleaningTimestamp);
            log.finest("holderTimeToLive: "      + holderTimeToLive     );
        }

        ArrayList toBeRemoved = null;
        SyncHolder h;
        Object key, value;
        Iterator i = null;

        synchronized (cache) {

            if ((now - lastCleaningTimestamp) <= CLEANING_PERIOD) {
                log.finest("No purging required");
                return false;
            }

            log.finest("Performing purging...");

            toBeRemoved = new ArrayList();

            i = cache.keySet().iterator();
            while (i.hasNext()) {
                key = i.next();
                value = cache.get(key);

                if (! (value instanceof SyncHolder)) {
                    // You shouldn't be here matey!!!
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Found unexpected object: " + key + " will be removed!");
                    }
                    toBeRemoved.add(key);
                    continue;
                }
                h = (SyncHolder)value;
                if ( (now - h.getCreationTimestamp()) > holderTimeToLive) {
                    log.finest("Purging holder for session " + key);
                    try {
                        h.close();
                    } catch (Exception e) {
                        log.severe(e.getMessage());
                        log.throwing(getClass().getName(), "clean", e);
                    }
                    toBeRemoved.add(key);
                }
            }

            //
            // Now remove the purged objects
            //
            i = toBeRemoved.iterator();
            while (i.hasNext()) {
                cache.remove(i.next());
            }
        }

        lastCleaningTimestamp = System.currentTimeMillis();

        return true;
    }
}
