/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.notification;

import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jibx.runtime.BindingDirectory;
import org.jibx.runtime.IBindingFactory;
import org.jibx.runtime.IMarshallingContext;

import com.funambol.framework.core.*;
import com.funambol.framework.core.dm.ddf.*;
import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerName;
import com.funambol.framework.notification.BootstrapMessageBuilder;
import com.funambol.framework.notification.NotificationException;
import com.funambol.framework.tools.ArrayUtils;
import com.funambol.framework.tools.Base64;
import com.funambol.framework.tools.DbgTools;
import com.funambol.framework.tools.WBXMLTools;


/**
 * It's a builder for a bootstrap message for Plain Profile.
 *
 *
 *
 * @version $Id: PlainBootstrapMessageBuilder.java,v 1.3 2006/08/07 21:09:25 nichele Exp $
 */
public class PlainBootstrapMessageBuilder implements BootstrapMessageBuilder {

    // --------------------------------------------------------------- Constants

    // -------------------------------------------------------------- Properties

    // ------------------------------------------------------------ Private data
    private Logger log = null;

    // ------------------------------------------------------------- Constructor
    public PlainBootstrapMessageBuilder() {
        log = FunambolLogger.getLogger(FunambolLoggerName.SERVER_DM_BOOTSTRAP);
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Returns the wbxml representation of the bootstrap message for the given <code>SycnMLDM</code> object.
     *
     * @param serverUri the server URI
     * @param deviceId the device id
     * @param phoneNumber the phone number
     * @param syncMLDM the SyncMLDM object
     *
     * @throws NotificationException if a error occurs during message creation
     * @return byte[] the wbxml representation of the bootstrap messege
     */
    public byte[] buildMessage(String serverUri, String deviceId,
                               String phoneNumber, SyncMLDM syncMLDM) throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("Calling buildMessage in PlainBootstrapMessageBuilder");
        }

        SyncML syncMLBootstrapMessage = getSyncMLBootstrapMessage(serverUri, deviceId, syncMLDM);

        byte[] message = null;

        try {

            ByteArrayOutputStream bout = new ByteArrayOutputStream();

            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IMarshallingContext c = f.createMarshallingContext();
            c.setIndent(0);
            c.marshalDocument(syncMLBootstrapMessage, "UTF-8", null, bout);

            String inputXml = new String(bout.toByteArray());
            inputXml = metInfNamespaceHandler(inputXml);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("PlainBootstrapMessage - xml: " + inputXml);
            }


            message = WBXMLTools.toWBXML(inputXml);

            /**
             * ONLY FOR TEST!
             */
            //writeBytesToFile("testwbxml.wbxml", message);

        } catch (Exception ex) {
            throw new NotificationException("Error during xml to wbxml transformation", ex);
        }

        return message;
    }

    // --------------------------------------------------------- Private Methods

    /**
     * Returns the <code>SyncML</code> object corresponds to the bootstrap message for the given <code>SycnMLDM</code> object.
     * @param serverUri the server uri
     * @param deviceId the device id
     * @param syncMLDM the SyncMLDM object
     *
     * @throws NotificationException if a error occurs during message creation
     * @return SyncML the <code>SyncML</code> corresponds to the bootstrap message
     */
    private SyncML getSyncMLBootstrapMessage(String serverUri, String deviceId, SyncMLDM syncMLDM) throws NotificationException {
        SyncML syncMLBootstrapMessage = null;

        SyncHdr syncHdrMessage = null;
        SyncBody syncBodyMessage = null;

        try {

            syncHdrMessage = new SyncHdr(
                new VerDTD("1.1"),
                Constants.SYNCML_DM_1_1,
                new SessionID("0"),
                "0", // MsgID
                new Target(deviceId),
                new Source(serverUri),
                null, /* response URI */
                false, /* no reponse   */
                null, /* credentials  */
                null
                /* metadata     */
                );

            Add addCommand = buildAddCommand(syncMLDM);

            syncBodyMessage = new SyncBody(
                new AbstractCommand[] {addCommand}
                ,
                true
                /* final */
                );

            syncMLBootstrapMessage = new SyncML(syncHdrMessage, syncBodyMessage);


        } catch (RepresentationException ex) {
            throw new NotificationException("Error during creation of the SyncML message", ex);
        }

        return syncMLBootstrapMessage;
    }


    /**
     * Returns the Add command for the parameters of the given <code>SyncMLDM</code>
     *
     * @param syncMLDM the <code>SyncMLDM</code> object
     * @return Add the add command for the parameters of the given <code>SyncMLDM</code>.
     * The CmdID is equal to 1.
     */
    private Add buildAddCommand(SyncMLDM syncMLDM) {
        Add addCommand = new Add(
            new CmdID(1),
            false, /* no response  */
            null, /* credential   */
            null,
            getItemsToAdd(syncMLDM)
        );

        return addCommand;
    }

    /**
     * Returns the <code>Item[]</code> for the parameters of the given <code>SyncMLDM</code>
     * @param syncMLDM the <code>SyncMLDM</code> object
     * @return Item[] the Item[] for the parameters of the given <code>SyncMLDM</code>.
     */
    private Item[] getItemsToAdd(SyncMLDM syncMLDM) {

        // item for DMAcc tree
        Item[] itemDMAcc = getItemsToAddForDMAcc(syncMLDM.getDmAcc());

        // item for Con tree
        Item[] itemCon = getItemsToAddForConNode(syncMLDM.getCon());

        return (Item[])(ArrayUtils.mergeArrays(itemDMAcc, itemCon, itemDMAcc.getClass().getComponentType()));
    }


    /**
     * Returns the <code>Item[]</code> for the parameters of the given <code>DMAcc</code>
     * @param dmAcc the <code>DMAcc</code> object
     * @return Item[] the Item[] for the parameters of the given <code>DMAcc</code>.
     */
    private Item[] getItemsToAddForDMAcc(DMAcc dmAcc) {

        Set keys = dmAcc.getDMAccounts().keySet();

        int numDMAcc = keys.size();

        if (numDMAcc == 0) {
            return new Item[0];
        }

        ArrayList listItems = new ArrayList();

        java.util.Map accounts = dmAcc.getDMAccounts();
        Iterator itAccounts    = keys.iterator();

        Meta metaInfoForInteriorNode = new Meta();
        metaInfoForInteriorNode.setFormat("node");

        DMAccount dmAccount = null;
        String accountName = null;
        while (itAccounts.hasNext()) {
            accountName = (String)itAccounts.next();
            dmAccount = (DMAccount)accounts.get(accountName);

            /*
             * Item for dm account
             */
            String localUri = SyncMLDM.SYNCML_DM_BASE_URI + SyncMLDM.SYNCML_DM_DMACC + "/" + accountName;
            Item itemDMAccount = new Item(new Target(localUri),
                                          null, /* Source */
                                          metaInfoForInteriorNode,
                                          null, /* ComplexData */
                                          false
                                          /* MoreData    */
                                 );

            listItems.add(itemDMAccount);

            /* Item for address */
            listItems.add(getItem(localUri + DMAccount.DMACC_ADDR, dmAccount.getAddress(), false, false));

            /* Item for addressType */
            listItems.add(getItem(localUri + DMAccount.DMACC_ADDR_TYPE,
                                  String.valueOf(dmAccount.getAddressType()), false, false));

            /* Item for port number */
            listItems.add(getItem(localUri + DMAccount.DMACC_PORT_NBR,
                                  String.valueOf(dmAccount.getPortNumber()), false, false));

            /* Item for conRef */
            String conRef = dmAccount.getConRef();
            if (conRef != null && !conRef.equals("")) {
                listItems.add(getItem(localUri + DMAccount.DMACC_CON_REF, conRef, false, false));
            }


            /* Item for serverId */
            listItems.add(getItem(localUri + DMAccount.DMACC_SERVER_ID, dmAccount.getServerId(), false, false));

            /* Item for serverPw */
            listItems.add(getItem(localUri + DMAccount.DMACC_SERVER_PASSWORD,
                                  dmAccount.getServerPassword(), false, false));

            /* Item for serverNonce */
            if (dmAccount.getServerNonce() != null) {
                String serverNonceB64 = new String(Base64.encode(dmAccount.getServerNonce()));
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("ServerNonce: " + DbgTools.bytesToHex(dmAccount.getServerNonce()));
                    log.finest("ServerNonceB64: " + serverNonceB64);

                }
                listItems.add(getItem(localUri + DMAccount.DMACC_SERVER_NONCE,
                                      serverNonceB64, false, false));
            }

            /* Item for userName */
            listItems.add(getItem(localUri + DMAccount.DMACC_USERNAME, dmAccount.getUserName(), false, false));

            /* Item for clientPw */
            listItems.add(getItem(localUri + DMAccount.DMACC_CLIENT_PASSWORD,
                                  dmAccount.getClientPassword(), false, false));

            /* Item for clientNonce */
            if (dmAccount.getClientNonce() != null) {
                String clientNonceB64 = new String(Base64.encode(dmAccount.getClientNonce()));
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("ClientNonce: " + DbgTools.bytesToHex(dmAccount.getClientNonce()));
                    log.finest("ClientNonceB64: " + clientNonceB64);
                }
                listItems.add(getItem(localUri + DMAccount.DMACC_CLIENT_NONCE,
                                      clientNonceB64, false, false));
            }

            /* Item for authPref */
            listItems.add(getItem(localUri + DMAccount.DMACC_AUTH_PREF, dmAccount.getAuthPref(), false, false));

            /* Item for name */
            listItems.add(getItem(localUri + DMAccount.DMACC_NAME, dmAccount.getName(), false, false));
        }

        return (Item[])listItems.toArray(new Item[0]);
    }


    /**
     * Returns the <code>Item[]</code> for the parameters of the given <code>ConNode</code>
     * @param conNode the <code>ConNode</code> object
     * @return Item[] the Item[] for the parameters of the given <code>ConNode</code>.
     */
    private Item[] getItemsToAddForConNode(ConNode conNode) {
        if (conNode == null) {
            return new Item[0];
        }
        Set keys = conNode.getConnections().keySet();
        if (keys.size() == 0) {
            return new Item[0];
        }

        ArrayList listItems = new ArrayList();
        Iterator itKeys = keys.iterator();

        Connection connection = null;
        String connectionName = null;
        while (itKeys.hasNext()) {
            connectionName = (String)itKeys.next();
            connection = conNode.getConnection(connectionName);
            listItems.addAll( getItemsToAddForConnection(connection, connectionName));
        }
        return (Item[])listItems.toArray(new Item[0]);
    }


    /**
     * Returns the <code>ArrayList</code> with the <code>Item</code> for the parameters of the given <code>Connection</code>
     * @param con the <code>Connection</code> object
     * @param connectionName the name of the connection
     * @return ArrayList the items for the parameters of the given <code>Connection</code>.
     */
    private ArrayList getItemsToAddForConnection(Connection con, String connectionName) {
       ArrayList listItems = new ArrayList();

       Meta metaInfoForInteriorNode = new Meta();
       metaInfoForInteriorNode.setFormat("node");

       /*
        * Item for connection node
        */
       String localUri = SyncMLDM.SYNCML_DM_BASE_URI + SyncMLDM.SYNCML_DM_CON + "/" + connectionName;
       Item itemConnection = new Item(new Target(localUri),
                                     null, /* Source */
                                     metaInfoForInteriorNode,
                                     null, /* ComplexData */
                                     false /* MoreData    */
                                     );

       listItems.add(itemConnection);

       com.funambol.framework.core.dm.ddf.Ext conExt = con.getExt();
       NAP conNap = con.getNap();
       PX  conPX  = con.getPX();

       ArrayList itemConExt = getItemsToAddForConExt(conExt, connectionName);
       ArrayList itemConNap = getItemsToAddForConNap(conNap, connectionName);
       ArrayList itemConPX = getItemsToAddForConPX(conPX, connectionName);

       listItems.addAll(itemConExt);
       listItems.addAll(itemConNap);
       listItems.addAll(itemConPX);

       return listItems;
    }

    /**
     * Returns the <code>ArrayList</code> with the <code>Item</code> for the parameters of the given <code>Ext</code>
     * @param ext the <code>Ext</code> object
     * @param connectionName the name of the connection of the given <code>Ext</code>
     * @return ArrayList the items for the parameters of the given <code>Ext</code>.
     */
    private ArrayList getItemsToAddForConExt(com.funambol.framework.core.dm.ddf.Ext ext, String connectionName) {
        ArrayList listItems = new ArrayList();

        if (ext == null) {
            return listItems;
        }

        java.util.Map nodes = ext.getNodes();
        int size = nodes.size();
        if (size == 0) {
            return listItems;
        }

        Meta metaInfoForInteriorNode = new Meta();
        metaInfoForInteriorNode.setFormat("node");

        /*
         * Item for Ext node
         */
        String localUri = SyncMLDM.SYNCML_DM_BASE_URI + SyncMLDM.SYNCML_DM_CON + "/" + connectionName + Connection.CON_EXT;
        Item itemConExt = new Item(new Target(localUri),
                                   null, /* Source */
                                   metaInfoForInteriorNode,
                                   new ComplexData(), /* ComplexData */
                                   false /* MoreData    */
                          );


        listItems.add(itemConExt);

        Iterator itNodes = nodes.keySet().iterator();

        String name = null;
        String value = null;
        while (itNodes.hasNext()) {
            name  = (String)itNodes.next();
            value = (String)nodes.get(name);
            listItems.add(getItem(localUri + "/" + name,
                            value, false, false));
        }

        return listItems;

    }

    /**
     * Returns the <code>ArrayList</code> with the <code>Item</code> for the parameters of the given <code>NAP</code>
     * @param nap the <code>NAP</code> object
     * @param connectionName the name of the connection of the given <code>NAP</code>
     * @return ArrayList the items for the parameters of the given <code>NAP</code>.
     */
    private ArrayList getItemsToAddForConNap(NAP nap, String connectionName) {
        ArrayList listItems = new ArrayList();

        if (nap == null) {
            return listItems;
        }


        Meta metaInfoForInteriorNode = new Meta();
        metaInfoForInteriorNode.setFormat("node");

        /*
         * Item for connection node
         */
        String localUri = SyncMLDM.SYNCML_DM_BASE_URI + SyncMLDM.SYNCML_DM_CON + "/" + connectionName + Connection.CON_NAP;
        Item itemConNap = new Item(new Target(localUri),
                                      null, /* Source */
                                      metaInfoForInteriorNode,
                                      new ComplexData(), /* ComplexData */
                                      false /* MoreData    */
                                      );


        listItems.add(itemConNap);

        /* Item for bearer */
        listItems.add(getItem(localUri + NAP.NAP_BEARER,
                              nap.getBearer(), false, false));

        /* Item for AddrType */
        listItems.add(getItem(localUri + NAP.NAP_ADDR_TYPE,
                              nap.getAddressType(), false, false));

        /* Item for Addr */
        listItems.add(getItem(localUri + NAP.NAP_ADDR, nap.getAddress(), false, false));


        if (nap.getAuths().size() > 0) {
            Item itemAuth = new Item(new Target(localUri + NAP.NAP_AUTH),
                                     null, /* Source */
                                     metaInfoForInteriorNode,
                                     null, /* ComplexData */
                                     false /* MoreData    */
                                     );

            listItems.add(itemAuth);

            Set keysAuths = nap.getAuths().keySet();
            Iterator itKeys = keysAuths.iterator();
            Auth conAuth = null;
            String authName = null;
            while (itKeys.hasNext()) {
                authName = (String)itKeys.next();
                conAuth = nap.getAuth(authName);
                listItems.addAll(getItemsToAddForConAuth(conAuth, localUri + NAP.NAP_AUTH + "/" + authName));
            }
        }

       return listItems;
    }

    /**
     * Returns the <code>ArrayList</code> with the <code>Item</code> for the parameters of the given <code>PX</code>
     * @param px the <code>PX</code> object
     * @param connectionName the name of the connection of the given <code>PX</code>
     * @return ArrayList the items for the parameters of the given <code>PX</code>.
     */
    private ArrayList getItemsToAddForConPX(PX px, String connectionName) {
        ArrayList listItems = new ArrayList();

        if (px == null) {
            return listItems;
        }


        Meta metaInfoForInteriorNode = new Meta();
        metaInfoForInteriorNode.setFormat("node");

        /*
         * Item for connection node
         */
        String localUri = SyncMLDM.SYNCML_DM_BASE_URI + SyncMLDM.SYNCML_DM_CON + "/" + connectionName + Connection.CON_PX;
        Item itemConPx = new Item(new Target(localUri),
                                      null, /* Source */
                                      metaInfoForInteriorNode,
                                      new ComplexData(), /* ComplexData */
                                      false /* MoreData    */
                                      );


        listItems.add(itemConPx);

        /* Item for PortNumber */
        listItems.add(getItem(localUri + PX.PX_PORT_NBR,
                              px.getPortNbr(), false, false));

        /* Item for AddrType */
        listItems.add(getItem(localUri + PX.PX_ADDR_TYPE,
                              px.getAddressType(), false, false));

        /* Item for Addr */
        listItems.add(getItem(localUri + PX.PX_ADDR, px.getAddress(), false, false));

        if (px.getAuths().size() > 0) {
            Item itemAuth = new Item(new Target(localUri + PX.PX_AUTH),
                                     null, /* Source */
                                     metaInfoForInteriorNode,
                                     null, /* ComplexData */
                                     false /* MoreData    */
                                     );

            listItems.add(itemAuth);

            Set keysAuths = px.getAuths().keySet();
            Iterator itKeys = keysAuths.iterator();
            Auth conAuth = null;
            String authName = null;
            while (itKeys.hasNext()) {
                authName = (String)itKeys.next();
                conAuth = px.getAuth(authName);
                listItems.addAll(getItemsToAddForConAuth(conAuth, localUri + PX.PX_AUTH + "/" + authName));
            }
        }

       return listItems;
    }


    /**
     * Returns the <code>ArrayList</code> with the <code>Item</code> for the parameters of the given <code>Auth</code>
     * @param auth the <code>Auth</code> object
     * @param localUri the base uri for the items
     * @return ArrayList the items for the parameters of the given <code>Auth</code>.
     */
    private ArrayList getItemsToAddForConAuth(Auth auth, String localUri) {
        ArrayList listItems = new ArrayList();

        Meta metaInfoForInteriorNode = new Meta();
        metaInfoForInteriorNode.setFormat("node");

        /*
         * Item for auth node
         */
        Item itemConnection = new Item(new Target(localUri),
                                      null, /* Source */
                                      metaInfoForInteriorNode,
                                      null, /* ComplexData */
                                      false /* MoreData    */
                                      );


        listItems.add(itemConnection);

        /* Item for id */
        listItems.add(getItem(localUri + Auth.AUTH_ID,
                              auth.getId(), false, false));

        /* Item for AddrType */
        listItems.add(getItem(localUri + Auth.AUTH_SECRET,
                              auth.getSecret(), false, false));


       return listItems;
    }


    /**
     * Returns a <code>Item</code> with the given localUri and data.
     *
     * @param localUri the local uri of the item
     * @param data the data of the item
     * @param interiorNode if <code>true</code> then the meta information is added to item in according
     *         with the SyncML DM standard object specification.
     * @return Item
     */
    private Item getItem(String localUri, String data, boolean interiorNode, boolean formatB64) {

        Meta meta = null;

        if (interiorNode) {
            meta = new Meta();
            meta.setFormat("node");
        }

        if (formatB64) {
            meta = new Meta();
            meta.setFormat("b64");
        }

        Item item = new Item(new Target(localUri),
                             null, /* Source */
                             meta,
                             new ComplexData(data),
                             false /* MoreData    */
                             );

        return item;
    }

    /**
     * This is a solution in order to obviate to a JiBX bug: it does
     * not allow to declare the namespace to level of structure.
     *
     * @param msg the server response
     *
     * @return the response with namespace correctly added into MetInf element
     */
    private String metInfNamespaceHandler(String msg) {
        int s = 0;
        int e = 0;

        StringBuffer response = new StringBuffer();
        while (( e = msg.indexOf("<Meta", s)) >= 0) {

            response = response.append(msg.substring(s, e));

            int a = msg.indexOf("</Meta>", e);
            String meta = msg.substring(e, a);

            meta = meta.replaceAll("<Type>"   , "<Type xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Format>" , "<Format xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Mark>"   , "<Mark xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Size>"   , "<Size xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Anchor>" , "<Anchor xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Version>", "<Version xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<NextNonce>" , "<NextNonce xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<MaxMsgSize>", "<MaxMsgSize xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<MaxObjSize>", "<MaxObjSize xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<EMI>"    , "<EMI xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Mem>"    , "<Mem xmlns='syncml:metinf'>");

            s = a + 7;
            response.append(meta).append("</Meta>");
        }
        return response.append(msg.substring(s, msg.length())).toString();
    }

    /**
     * THIS METHOD IS USES ONLY FOR TEST!!!
     * <p>Writes the given byte[] in a file with the given fileName.
     * @param fileName the name of the file
     * @param bytes the byte[] to write
     */
    private void writeBytesToFile(String fileName, byte[] bytes) {
        try {
            File file = new File(fileName);
            FileOutputStream fOut = new FileOutputStream(file);
            fOut.write(bytes);
        } catch (FileNotFoundException ex) {
        } catch (IOException ex) {
        }

    }
}
